#!/bin/bash

#
# Cleanse files for waiving purposes
#
# This module provides utilities to employ "cleansing" mechanism, which
# is a method to dynamically control known issues without limiting
# effectivity of test suites.
#
#
# =head2 THE "CLEANSING" METHOD
#
# Automated tests are often best implemented in a way that they compile
# list of unexpected conditions (eg. error log lines, bad words,...)
# and then they verify that the list is empty.
#
#     grep -w ERROR /var/log/foo.log >foo.log.errors
#     jat__cmd -S 1 -h "no errors in foo.log" \
#         test -s foo.log.errors
#
# If for such test, an error condition appears, the list is not empty
# anymore and test fails.
#
# This is all fine & dandy, but there is a problem:  Between the moment
# the issue is detected and fixed, there inevitably be a window where
# the same issue will be detected over and over, rendering the test
# useless.  Furthermore, in cases where the issue has low priority,
# (or even worse, is a false positive) the window may end up
# extraordinarily large: it can take days, months or years before
# a small issue is really fixed.
#
# The "cleansing" method attacks this problem is by inserting one extra
# step between the generating of the issue list, and the emptiness
# check:
#
#     grep -w ERROR /var/log/foo.log >foo.log.errors
#     cleanse foo.log.errors >foo.log.errors.cleansed
#     jat__cmd -S 1 -h "no errors in foo.log" \
#         test -s foo.log.errors.cleansed
#
# This way, whenever there is an issue that won't be solved immediately,
# the 'cleanse' command can remove it from the log, so that it won't be
# detected over and over.
#
# The way this module fits in is that it provides you with utilities
# to create the 'cleanse' commands that are dynamic, flexible and
# transparent to the log reader.  That is, it lets you implement own
# generator for the cleanser, which it then runs on the given list,
# automatically creating and submitting all the relics: original file,
# cleansed file, their diff, and the cleanser script itself.
#
# Generating the cleanser script dynamically enables you to select
# for the right conditions where it's relevant to actually remove
# the issue.  For example, you might want to remove an error only
# on `ppc64` architecture, perhaps because you know it it's caused
# by a bug that only reproduces on this architecture; seeing the
# error anywhere else is still a valid detection.
#
# Furthermore, it allows you to use comments to link given exceptions
# to the filed issues, and pruning the list of issues from time
# to time.
#
# Finally, the architecture of this module supports having only one
# cleanser generator for multiple files; this makes it easier to
# keep all the exceptions in one place.
#
# =head2 EXAMPLE
#
# For example, a cleanser generator can look like this:
#
#     _cleanse__mkcleanser() {
#         local file=$1
#         echo '!#/usr/bin/sed -f'
#         case $file in
#             foo.log)
#                 echo '/foo/d'
#                 # ^^ https://tracker.foo.bar/issue/1234
#                 if test "$(arch)" == "ppc64"; then
#                     echo '/blah/d'
#                 fi
#                 # ^^ https://tracker.foo.bar/issue/5678
#                 if such_and_such_scenario; then
#                     echo '/hello/d'
#                 fi
#                 # ^^ https://tracker.foo.bar/issue/8901
#                 ;;
#             *) jat__log_warning "unknown FILE, leaving data intact: $file"
#         esac
#     }
#
# By letting the cleanser script be defined dynamically from the
# test code, we make it easier to apply the exceptions only when
# needed (5678 and 8901), use own test framework to determine such
# conditions (8901), and allow us to mark the exceptions with
# trackable links so that we can always come back and weed out the
# old ones (1234, 5678, 8901).
#

#
# Disable cleansers?
#
# If 'true', cleansers in cleanse__file() will have no effect.  Enable
# this if you want to see which of your tests will fail without the
# cleansing in place effect, perhaps to review and remove old cleansers.
#
CLEANSE__DISABLE=${CLEANSE__DISABLE:-false}

#
# Cleanser generator command
#
# Command to dynamically generate cleanser.  Must accept name of the
# file being cleansed and output a cleanser script (including shebang
# line!).
#
# The cleanser script must act as a UNIX filter, ie. accept content on
# standard input, and produce cleansed version of the content on standard
# output.  The script can be written in any language, although `sed`, `awk`
# or `perl` are especially suitable.
#
CLEANSE__GENERATOR=${CLEANSE__GENERATOR:-cleanse__mkcleanser}

#
# Original file name inside cleanser scripts
#
# This variable is set only inside cleanser generator and inside cleanser
# itself and contains filename of the file being cleansed.  It's equivalent
# to first argument of the cleanser generator.
#
CLEANSE__ORIG_NAME=${CLEANSE__ORIG_NAME:-}

#
# Original file path inside cleanser scripts
#
# This variable is set only inside cleanser generator and inside cleanser
# itself and contains full path to the file being cleansed.
#
CLEANSE__ORIG_PATH=${CLEANSE__ORIG_PATH:-}

cleanse__file() {
    #
    # Cleanse files $1 based on cleanser of the same name
    #
    # Usage:
    #     cleanse__file [options] FILE
    #
    # FILE will be cleansed using command provided as value of
    # $CLEANSE__GENERATOR.
    #
    # NOTE: FILE may be a full path, but the cleansing is only based on
    #       the filename!  That is, following calls will use the same
    #       cleansing code:
    #
    #           cleanse__file /tmp/123/foo.txt
    #           cleanse__file /tmp/456/foo.txt
    #
    #       just with different files.
    #
    # Resulting file name is formed by appending suffix to the original.
    # The default suffix is `.cleansed` and can be changed by passing
    # option -s followed by new suffix.
    #
    # Using -c option, you can override cleanser command supplied
    # as CLEANSE__GENERATOR envvar; the option will be preferred.
    #
    # By default, the function submits (using jat__submit) original
    # file, cleansed file, unified diff between these two, and the
    # used cleanser script.  Options -O, -C, -D and -S can be
    # used to disable any of the four, respectively.
    #
    # See _cleanse__mkcleanser() for more details.
    #
    local orig                  # original file
    local orign                 # ^^^ .. name
    local suffix=.cleansed      # suffix to form resulting file
    local s_orig=true           # submit original?
    local s_clnd=true           # submit cleansed?
    local s_diff=true           # submit diff?
    local s_clnr=true           # submit cleanser?
    local cmaker=$CLEANSE__GENERATOR   # cleanser command
    while true; do case $1 in
        -C) s_clnd=false; shift ;;
        -D) s_diff=false; shift ;;
        -O) s_orig=false; shift ;;
        -S) s_clnr=false; shift ;;
        -s) suffix=$2; shift 2 ;;
        -c) cmaker=$2; shift 2 ;;
        --) shift; break ;;
        *)  break ;;
    esac done
    type -t "$cmaker" >/dev/null || {
        jat__log_error \
            ".=================================--"                        \
            "| cleanser maker command not found: $cmaker"                 \
            "|---------------------------------------------------------." \
            "| Implement it under default name or use -c to specify    |" \
            "| different command.  The command must accept path to     |" \
            "| cleansed file and print a cleansing script for that     |" \
            "| file.                                                   |" \
            "'---------------------------------------------------------'"
        return 2
    }
    orig=$1
    test -f "$orig" || {
        jat__log_error "no such file: $orig"
        return 3
    }
    orign=$(basename "$orig")
    tmp=$(mktemp -dt cleanse__file.XXXXXXXX)
    if $CLEANSE__DISABLE; then
        printf '%s\n' "#!/bin/sh" "cat" >"$tmp/cleanser"
    else
        CLEANSE__ORIG_PATH="$orig" \
        CLEANSE__ORIG_NAME="$orign" \
            "$cmaker" "$orign" > "$tmp/cleanser"
    fi
    chmod +x "$tmp/cleanser"
    jat__log_info "cleansing file: $orign to $orign$suffix"
    #shellcheck disable=SC2094
    <"$orig" CLEANSE__ORIG_PATH="$orig" CLEANSE__ORIG_NAME="$orign" "$tmp/cleanser" >"$orig$suffix" || {
        jat__log_error "cleanser exited with non-zero status"
        return 3
    }
    diff -u "$orig" "$orig$suffix" >"$tmp/$orign$suffix.diff"
    $s_orig && jat__submit "$orig" "$orign"
    $s_clnd && jat__submit "$orig$suffix" "$orign$suffix"
    $s_clnr && jat__submit "$tmp/cleanser" "$orign.cleanser"
    $s_diff && jat__submit "$tmp/$orign$suffix.diff" "$orign$suffix.diff"
    rm -r "$tmp"
}

#          #                                your cursor will be cleansed #
# INTERNAL # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
#          #                              the moment you cross this line #


#           #                                                            #
# TEMPLATES # ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
#           #                                                            #

_cleanse__mkcleanser() {
    #
    # Print cleanser named $1
    #
    # This function is basically data container to help keep various
    # false positives and known issues at bay.  The output is a sed script
    # that can be used to filter out a log or similar relic before
    # assessing whether it contains errors worth reporting.
    #
    # Main reason these cleansers are kept under library function is that
    # it allows applying some dynamic.  Similar or even same log lines may
    # be OK on one scenario but not another, or under one architecture
    # but not the other, so this allows printing the smallest possible
    # set of exceptions.
    #
    # The other reason is that it allows us to link exceptions to bugs
    # using comments in this function, which then allows us to review
    # the list from time to time and remove obsolete exceptions.
    #
    #shellcheck disable=SC2016
    case $CLEANSE__ORIG_NAME in
        foo.log)
            echo '#!/usr/bin/sed -f'
            #https://tracker.foo.bar/issue/1234
            echo '/foo/d'
            ;;
        *) jat__log_warning "unknown FILE, leaving data intact: $CLEANSE__ORIG_NAME"
    esac
}

#shellfu module-version=0.0.5
